from yta_editor_frame.frame.metadata import _FrameMetadata, _FrameMetadataExtractor
from yta_editor_frame.frame.resource.identifier import FrameResourceIdentifier
from yta_editor_frame.frame.resource.converter import FrameResourceConverter
from yta_editor_frame.frame.resource.types import FrameResourceType
from yta_validation.parameter import ParameterValidator
from yta_validation import PythonValidator
from typing import Union


# TODO: Maybe rename to 'VideoFrame' or similar, but
# careful with the 'pyav.VideoFrame' then...
class Frame:
    """
    A frame, the main class that is shared in between
    the nodes and all the rendering process.

    If the `metadata` is provided, it will be set. If
    it is not provided, it will be extracted if the
    `do_autoextract_metadata` is `True`, or set as
    unknown if `False`.
    """

    @property
    def size(
        self
    ) -> tuple[int, int]:
        """
        The size of the resource expressed as
        `(width, height)`, in pixels.
        """
        self._assert_alive()
        return (self.resource.width, self.resource.height)
    
    @property
    def render_context(
        self
    ) -> 'RenderContext':
        """
        The `RenderContext` instance that allows us to use
        the `RenderTargetProvider` and more resource 
        handlers.
        """
        return self.resource.render_context

    def __init__(
        self,
        # TODO: Set the 4 types we have
        resource,
        metadata: Union[dict, None] = None,
        do_autoextract_metadata: bool = False
    ):
        ParameterValidator.validate_dict('metadata', metadata)

        # TODO: Set the 4 types we have
        self.resource = resource
        self._ref_count: int = 1
        """
        *For internal use only*

        Internal counter to determine when no one is using
        this frame and we can destroy it and free the
        resources.
        """

        metadata = (
            {}
            if metadata == None else
            metadata
        )

        self.metadata = (
            self._get_metadata_autodetected()
            if do_autoextract_metadata else
            _FrameMetadata(**metadata)
        )
        """
        Metadata we want to include about the frame.
        """

    @classmethod
    def from_frame(
        cls,
        frame: Union['Frame', 'av.VideoFrame', 'np.ndarray', 'moderngl.Texture', 'RenderTarget'],
        render_context: 'RenderContext',
        metadata: dict = {}
    ) -> 'Frame':
        """
        Get a `Frame` instance by creating the resource from
        the `frame` parameter provided, that must be one of
        the following:
        - `av.VideoFrame`
        - `np.ndarray`
        - `moderngl.Texture`
        - `RenderTarget`
        """
        # TODO: Refactor if possible (?)
        from yta_editor_frame.frame.resource.av_video_frame_resource import AvVideoFrameResource
        from yta_editor_frame.frame.resource.cpu_frame_resource import CPUNumpyArrayFrameResource
        from yta_editor_frame.frame.resource.gpu_frame_resource import GPUTextureFrameResource
        from yta_editor_frame.frame.resource.render_target_frame_resource import RenderTargetFrameResource

        ParameterValidator.validate_mandatory_instance_of('frame', frame, ['Frame', 'VideoFrame', 'ndarray', 'Texture', 'RenderTarget'])
        ParameterValidator.validate_mandatory_instance_of('render_context', render_context, 'RenderContext')
        ParameterValidator.validate_mandatory_dict('metadata', metadata)

        if PythonValidator.is_instance_of(frame, 'Frame'):
            # TODO: Maybe we need to clone the resource...
            # return frame.retain()
            raise Exception('Using "from_frame" on a "Frame"... Maybe parse, mate (?)')
        
        resource = (
            AvVideoFrameResource(
                frame = frame,
                render_context = render_context
            )
            if PythonValidator.is_instance_of(frame, 'VideoFrame') else
            CPUNumpyArrayFrameResource(
                frame = frame,
                render_context = render_context
            )
            if PythonValidator.is_instance_of(frame, 'ndarray') else
            GPUTextureFrameResource(
                frame = frame,
                render_context = render_context
            )
            if PythonValidator.is_instance_of(frame, 'Texture') else
            RenderTargetFrameResource(
                frame = frame,
                render_context = render_context
            )
            # if PythonValidator.is_instance_of('RenderTarget') else
        )

        frame = Frame(
            resource = resource,
            metadata = metadata
            # TODO: Maybe send 'do_autoextract_metadata' as True (?)
        )

        return frame

    def _get_metadata_autodetected(
        self
    ) -> '_FrameMetadata':
        """
        *For internal use only*

        Autodetect the metadata from the resource.
        """
        return _FrameMetadataExtractor.extract_from_video(self.resource.frame)
    
    def _to_resource(
        self,
        type: FrameResourceType
    ) -> 'Frame':
        """
        *For internal use only*

        Get the `FrameResource` of the given `type`
        that must be associated to a new `Frame`
        instance.

        This method is meant to be used for the `_to()`
        and the `_as()` methods.
        """
        type = FrameResourceType.to_enum(type)

        if type not in FrameResourceType.get_all():
            raise Exception(f'The "type" {type.value} is not supported for the transformation.')
        
        if self.resource.type == type:
            return self.resource.copy()
        
        frame_resource_converter = FrameResourceConverter(self.render_context)

        return (
            frame_resource_converter.to_numpy(
                resource = self.resource
            )
            if type == FrameResourceType.CPU else
            frame_resource_converter.to_texture(
                resource = self.resource
            )
            if type == FrameResourceType.GPU else
            frame_resource_converter.to_videoframe(
                resource = self.resource
            )
            if type == FrameResourceType.CPU_AV_VIDEOFRAME else
            frame_resource_converter.to_render_target(
                resource = self.resource
            )
        )

    def _to(
        self,
        type: FrameResourceType
    ) -> 'Frame':
        """
        *For internal use only*

        Get a new instance with the same metadata but
        of the `type` provided, that must be one of
        the following:
        - `FrameResourceType.GPU`
        - `FrameResourceType.GPU_RENDER_TARGET`
        - `FrameResourceType.CPU`
        - `FrameResourceType.CPU_AV_VIDEOFRAME`

        This method will NOT release the resources of
        the previous instance.
        """
        type = FrameResourceType.to_enum(type)

        if type not in FrameResourceType.get_all():
            raise Exception(f'The "type" {type.value} is not supported for the transformation.')
        
        if self.resource.type == type:
            # TODO: Is this ok? Or should I do nothing (?)
            return self.retain()
        
        new_resource = self._to_resource(type)
        
        """
        By now we are returning a new instance to
        avoid errors if this `Frame` instance is
        shared and being used by other instances.
        """
        frame = Frame(
            resource = new_resource,
            metadata = self.metadata.to_dict()
        )
        
        return frame
    
    def to_texture(
        self
    ) -> 'Frame':
        """
        Get a new instance with the same metadata but
        of the `FrameResourceType.GPU`, that is the
        one that works with `moderngl.Texture`.

        This method will NOT release the resources of
        the previous instance.
        """
        self._assert_alive()

        return self._to(type = FrameResourceType.GPU)
    
    def to_render_target(
        self
    ) -> 'Frame':
        """
        Get a new instance with the same metadata but
        of the `FrameResourceType.GPU_RENDER_TARGET`,
        that is the one that works with `RenderTarget`.

        This method will NOT release the resources of
        the previous instance.
        """
        self._assert_alive()

        return self._to(type = FrameResourceType.GPU_RENDER_TARGET)
    
    def to_numpy(
        self
    ) -> 'Frame':
        """
        Get a new instance with the same metadata but
        of the `FrameResourceType.CPU`, that is the
        one that works with `np.ndarray`.

        This method will NOT release the resources of
        the previous instance.
        """
        self._assert_alive()

        return self._to(type = FrameResourceType.CPU)
    
    def to_videoframe(
        self
    ) -> 'Frame':
        """
        Get a new instance with the same metadata but
        of the `FrameResourceType.CPU_AV_VIDEOFRAME`,
        that is the one that works with `av.VideoFrame`.

        This method will NOT release the resources of
        the previous instance.
        """
        self._assert_alive()

        return self._to(type = FrameResourceType.CPU_AV_VIDEOFRAME)

    def _as(
        self,
        type: FrameResourceType
    ) -> 'Frame':
        """
        Mutate this Frame by converting its resource type.

        Only allowed if this Frame has exclusive ownership
        of its resource (ref_count == 1).
        """
        type = FrameResourceType.to_enum(type)

        if type not in FrameResourceType.get_all():
            raise Exception(f'The "type" {type.value} is not supported.')

        # No-op
        if self.resource.type == type:
            return self

        if self._ref_count != 1:
            raise Exception(
                f'Cannot mutate Frame with multiple references (references: {str(self._ref_count)}). '
                'Use `.to()` instead.'
            )

        old_resource = self.resource

        new_resource = self._to_resource(type)

        # Swap resource
        self.resource = new_resource

        # Release old AFTER swap
        # TODO: This is failing, I don't know why
        # old_resource.release()

        return self

    def as_numpy(
        self
    ) -> 'Frame':
        """
        Modify the current instance to transform it into
        a `FrameResourceType.CPU`, that is the one that
        works with the `np.ndarray`.
        
        This method will release the resources of the
        previous instance.
        """
        return self._as(FrameResourceType.CPU)
    
    def as_texture(
        self
    ) -> 'Frame':
        """
        Modify the current instance to transform it into
        a `FrameResourceType.GPU`, that is the one that
        works with the `moderngl.Texture`.

        This method will release the resources of the
        previous instance.
        """
        return self._as(FrameResourceType.GPU)
    
    def as_render_target(
        self
    ) -> 'Frame':
        """
        Modify the current instance to transform it into
        a `FrameResourceType.GPU_RENDER_TARGET`, that is
        the one that works with the `RenderTarget`.

        This method will release the resources of the
        previous instance.
        """
        return self._as(FrameResourceType.GPU_RENDER_TARGET)
    
    def as_videoframe(
        self
    ) -> 'Frame':
        """
        Modify the current instance to transform it into
        a `FrameResourceType.CPU_AV_VIDEOFRAME`, that is
        the one that works with the `av.VideoFrame`.

        This method will release the resources of the
        previous instance.
        """
        return self._as(FrameResourceType.CPU_AV_VIDEOFRAME)
    
    # TODO: Recommended by ChatGPT but still don't know...
    # def convert(
    #     self,
    #     type: FrameResourceType
    # ) -> 'Frame':
    #     """
    #     TODO: Add doc
    #     """
    #     if self._ref_count == 1:
    #         return self._as(type)
    #     else:
    #         return self._to(type)

    def retain(
        self
    ) -> 'Frame':
        """
        Method to indicate that the frame is being used by
        someone else so it must be retained and not sent
        back to the pool.
        """
        if self._ref_count <= 0:
            raise Exception('Cannot retain destroyed frame.')

        self._ref_count += 1

        return self

    def release(
        self
    ):
        """
        Method to indicate that the frame has been freed
        by one owner and must be completely released.

        If no one is owning this `Frame` instance, it
        will be destroyed, and if the instance needs to
        free the resource, it will be done.
        """
        self._ref_count -= 1

        if self._ref_count < 0:
            raise Exception('Frame released too many times.')

        if self._ref_count == 0:
            self._destroy()

    def _destroy(
        self
    ):
        """
        Method to destroy the frame instance, that will
        send the resource to the pool if it is a
        resource that must be released.
        """
        if self.resource:
            self.resource.release()

        self.resource = None

    def _assert_alive(
        self
    ):
        if self.resource is None:
            raise Exception('Frame already destroyed')

    def __repr__(
        self
    ):
        return f'<Frame rc={self._ref_count} res={self.resource}>'
    
# TODO: Implement
class AudioFrame:
    pass


from yta_constants.enum import YTAEnum as Enum


class VideoFrameMetadataField(Enum):
    """
    Metadata fields of a video `Frame` instance.
    """

    HAS_ALPHA_PIXELS = 'has_alpha_pixels'
    """
    Boolean field to indicate if the video `Frame`
    has transparent pixels or not, which means that
    include alpha channel.
    """
    IS_FULLY_OPAQUE = 'is_fully_opaque'
    """
    Boolean field to indicate if all the pixels of
    the video `Frame` are opaque (don't have alpha
    channel or have alpha channel as non-transparent).

    This field is useful when we are combining frames
    and we find one like this, so we can stop
    combining the next one because this one, being
    completely opaque, will cover the following ones.
    """
    IS_FULLY_TRANSPARENT = 'is_fully_transparent'
    """
    Boolean field to indicate if all the pixels of
    the video `Frame` are transparent (have alpha
    channel and are completely transparent).)

    This field is useful when we are combining frames
    and we find one like this, so we can directly omit
    it and go to the next one because this one, being
    completely transparent, will let the following one
    be seen.
    """
    IS_FROM_GAP = 'is_from_gap'
    """
    Boolean field to identify a frame that comes from
    a gap track item, which means that there is nothing
    placed in that position and it must be completely
    black or transparent (depending on the other
    elements in the timeline for that same exact moment).
    """

class AudioFrameMetadataField(Enum):
    """
    Metadata fields of an audio `Frame` instance.
    """

    IS_SILENT = 'is_silent'
    """
    Boolean field to indicate if the audio frame is
    silent, which means that there is no audio to
    play and can be omitted.
    """


from yta_editor_frame.frame.metadata.fields import AudioFrameMetadataField, VideoFrameMetadataField
from yta_validation import PythonValidator
from dataclasses import dataclass
from typing import Union


@dataclass
class _FrameMetadata:
    """
    *For internal use only*

    Class to wrap the metadata about an audio or video
    frame. Any value that is `None` means that the data
    is unknown.

    The metadata is useful to know if the frame has any
    kind of transparency, if it is fully transparent or
    fully opaque, if the audio frame is silent, etc.

    This class uses an internal dict to manage the
    information and works directly with it. Only the
    fields that are registered in the following classes
    are accepted:
    - `AudioFrameMetadataField`
    - `VideoFrameMetadataField`
    """

    def __init__(
        self,
        **kwargs
    ):
        object.__setattr__(self, '_metadata', {})
        """
        *For internal use only*

        The actual metadata dict.
        """

        for key, value in kwargs.items():
            self.__setattr__(key, value)

    def copy(
        self
    ):
        """
        Create a copy of this metadata by shallow copying
        the internal dictionary.
        """
        return _FrameMetadata(
            metadata = self.to_dict()
        )
    
    def to_dict(
        self
    ) -> dict:
        """
        Get a copy of the metadata directly as a dict.
        """
        return self._metadata.copy()
    
    def __setattr__(
        self,
        name: str,
        value: any
    ):
        if (
            not VideoFrameMetadataField.is_valid_value(name) and
            not AudioFrameMetadataField.is_valid_value(name)
        ):
            raise Exception(f'The attribute "{name}" is not accepted in our metadata.')
        
        if self._is_internal_attribute(name):
            object.__setattr__(self, name, value)
        else:
            self._metadata[name] = value

    def __getattr__(
        self,
        name: str
    ) -> Union[any, None]:
        return self._metadata.get(name, None)

    def __delattr__(
        self,
        name: str
    ):
        if self._is_internal_attribute(name):
            object.__delattr__(self, name)
        else:
            try:
                del self._metadata[name]
            except KeyError:
                raise AttributeError(f'{name} not found')
            
    def _is_internal_attribute(
        self,
        name: str
    ) -> bool:
        """
        Decide whether an attribute belongs to Python/runtime
        or should be treated as metadata.
        """
        return (
            name.startswith('_') or
            hasattr(type(self), name)
        )
    
    def copy(
        self
    ):
        """
        Get a new `_FrameMetadata` instance with a copy
        of the metadata inside.
        """
        frame_metadata = _FrameMetadata()
        object.__setattr__(frame_metadata, '_metadata', self._metadata.copy())

        return frame_metadata

    def update(
        self,
        dict: dict
    ):
        """
        Update the internal metadata of this instance.
        """
        self._metadata.update(dict)

    def __repr__(
        self
    ):
        return f'FrameMetadata({self._metadata})'        

class _FrameMetadataExtractor:
    """
    Class to extract the information of a frame
    to a `_FrameMetadata` instance.
    """

    @staticmethod
    def extract_from_video(
        frame: Union['np.ndarray', 'moderngl.Texture', 'av.VideoFrame', 'RenderTarget']
    ) -> _FrameMetadata:
        """
        Extract the metadata from the given video
        `frame` that is easy to get and available.
        
        The `frame` must be one of the following types:
        - `np.ndarray`
        - `moderngl.Texture`
        - `av.VideoFrame`
        - `RenderTarget`
        
        There is metadata that has to be manually
        written when the frame is modified by a node
        and we know exactly some metadata of the
        output we will generate and return, but this
        is just an easy and automated extractor.
        """
        metadata_dict = {
            VideoFrameMetadataField.HAS_ALPHA_PIXELS.value: None,
            VideoFrameMetadataField.IS_FULLY_OPAQUE.value: None,
            VideoFrameMetadataField.IS_FULLY_TRANSPARENT.value: None
        }

        # TODO: Extract it
        if PythonValidator.is_numpy_array(frame):
            # [1] `np.ndarray`
            if (
                frame.ndim == 3 and
                frame.shape[2] == 4
            ):
                alpha = frame[..., 3]
                a_min = alpha.min()
                a_max = alpha.max()
                metadata_dict[VideoFrameMetadataField.HAS_ALPHA_PIXELS.value] = a_min < 255
                metadata_dict[VideoFrameMetadataField.IS_FULLY_OPAQUE.value] = a_min == 255
                metadata_dict[VideoFrameMetadataField.IS_FULLY_TRANSPARENT.value] = a_max == 0
            else:
                metadata_dict[VideoFrameMetadataField.HAS_ALPHA_PIXELS.value] = False
                metadata_dict[VideoFrameMetadataField.IS_FULLY_OPAQUE.value] = True
                metadata_dict[VideoFrameMetadataField.IS_FULLY_TRANSPARENT.value] = False
        elif PythonValidator.is_instance_of(frame, 'VideoFrame'):
            # [2] `av.VideoFrame`
            metadata_dict[VideoFrameMetadataField.HAS_ALPHA_PIXELS.value] = 'a' in frame.format.name
            metadata_dict[VideoFrameMetadataField.IS_FULLY_OPAQUE.value] = (
                True
                if not metadata_dict[VideoFrameMetadataField.HAS_ALPHA_PIXELS.value] else
                # We don't know yet
                None
            )
            metadata_dict[VideoFrameMetadataField.IS_FULLY_TRANSPARENT.value] = (
                False
                if not metadata_dict[VideoFrameMetadataField.HAS_ALPHA_PIXELS.value] else
                # We don't know yet
                None
            )

        """
        # [3] `moderngl.Texture`
        # [4] `RenderTarget`
        
        The `moderngl.Texture` (or `RenderTarget`, that has
        a texture inside) are not easy to access, so we
        don't. If we are using a specific node that we know
        that will generate no transparency, then we can set
        that metadata manually, but this is to extract the
        information automatically.
        """

        return _FrameMetadata(
            **metadata_dict
        )
    
    def extract_from_audio(
        frame: 'av.AudioFrame'
    ) -> _FrameMetadata:
        """
        Extract the metadata from the given audio
        `frame` that is easy to get and available.
        
        The `frame` must be one of the following types:
        - `av.AudioFrame`
        
        There is metadata that has to be manually
        written when the frame is modified by a node
        and we know exactly some metadata of the
        output we will generate and return, but this
        is just an easy and automated extractor.
        """
        metadata_dict = {
            AudioFrameMetadataField.IS_SILENT.value: None
        }

        # TODO: Handle whatever I can

        return _FrameMetadata(
            **metadata_dict
        )


from yta_editor_frame.frame.resource.abstract import FrameResourceABC
from yta_editor_frame.frame.resource.types import FrameResourceType
from yta_validation.parameter import ParameterValidator


class CPUNumpyArrayFrameResource(
    FrameResourceABC
):
    """
    A frame resource based on a numpy array.

    The `.frame` attribute is a `np.ndarray`.
    """
    
    @property
    def width(
        self
    ) -> int:
        return self.frame.shape[1]
    
    @property
    def height(
        self
    ) -> int:
        return self.frame.shape[0]
    
    @property
    def format(
        self
    ) -> str:
        return str(self.frame.dtype)
    
    @property
    def type(
        self
    ) -> FrameResourceType:
        return FrameResourceType.CPU

    def __init__(
        self,
        frame: 'np.ndarray',
        render_context: 'RenderContext'
    ):
        ParameterValidator.validate_mandatory_numpy_array('frame', frame)

        super().__init__(
            frame = frame,
            render_context = render_context
        )

    def release(
        self
    ):
        pass

    def copy(
        self
    ) -> 'CPUNumpyArrayFrameResource':
        """
        Get a copy of this frame resource.
        """
        copy = self.frame.copy()

        return CPUNumpyArrayFrameResource(
            frame = copy,
            render_context = self.render_context
        )


from yta_editor_frame.frame.resource.abstract import FrameResourceABC
from yta_editor_frame.frame.resource.types import FrameResourceType
from yta_editor_pools.render_target import RenderTarget
from yta_editor_frame.utils import copy_texture
from yta_validation.parameter import ParameterValidator


class RenderTargetFrameResource(
    FrameResourceABC
):
    """
    A frame resource based on a render target.

    The `.frame` attribute is a `RenderTarget`.
    """
    
    @property
    def width(
        self
    ) -> int:
        return self.frame.width
    
    @property
    def height(
        self
    ) -> int:
        return self.frame.height
    
    @property
    def format(
        self
    ) -> str:
        return self.frame.dtype
    
    @property
    def type(
        self
    ) -> FrameResourceType:
        return FrameResourceType.GPU_RENDER_TARGET

    def __init__(
        self,
        frame: 'RenderTarget',
        render_context: 'RenderContext'
    ):
        ParameterValidator.validate_mandatory_instance_of('frame', frame, 'RenderTarget')

        super().__init__(
            frame = frame,
            render_context = render_context
        )

    def release(
        self
    ):
        if self.frame.is_owned:
            self.render_context.gpu.render_target.release(self.frame)
        else:
            # TODO: Maybe comment the exception (?)
            raise Exception('This RenderTarget is not owned and will be automatically released by its owner, not you.')
            pass

    def copy(
        self
    ) -> 'RenderTargetFrameResource':
        """
        Get a copy of this frame resource.
        """
        texture_copy = copy_texture(
            texture = self.frame.texture,
            render_context = self.render_context,
            do_get_from_pool = True
        )

        copy = RenderTarget(
            context = self.render_context,
            texture = texture_copy,
            is_owned = self.frame.is_owned
        )

        return RenderTargetFrameResource(
            frame = copy,
            render_context = self.render_context
        )


"""
Module to transform one resource into another
one to be able to do CPU -> GPU, GPU -> CPU,
etc.
"""
from yta_editor_frame.frame.resource.identifier import FrameResourceIdentifier
from yta_editor_frame.frame.resource.cpu_frame_resource import CPUNumpyArrayFrameResource
from yta_editor_frame.frame.resource.gpu_frame_resource import GPUTextureFrameResource
from yta_editor_frame.frame.resource.av_video_frame_resource import AvVideoFrameResource
from yta_editor_frame.frame.resource.render_target_frame_resource import RenderTargetFrameResource
from yta_editor_frame.utils import copy_texture
from yta_programming.decorators.requires_dependency import requires_dependency
from typing import Union


FORMAT = 'rgba'

class FrameResourceConverter:
    """
    Class able to transform frames from one type to
    another.

    Singleton class for a specific `RenderContext`.
    """

    _instances: dict[int, 'FrameResourceConverter'] = {}

    @classmethod
    def get(
        cls,
        render_context: 'RenderContext'
    ) -> 'FrameResourceConverter':
        """
        Get the singleton instance for the specific
        `render_context` provided as parameter.
        """
        key = id(render_context)
        if key not in cls._instances:
            cls._instances[key] = cls(render_context)

        return cls._instances[key]
    
    def __init__(
        self,
        render_context: 'RenderContext'
    ):
        self.render_context: 'RenderContext' = render_context
        """
        The `RenderContext` associated to it.
        """

    @requires_dependency('numpy', 'yta_editor_frame', 'numpy')
    def to_numpy(
        self,
        resource: Union[CPUNumpyArrayFrameResource, GPUTextureFrameResource]
    ) -> CPUNumpyArrayFrameResource:
        """
        Transform the `resource` provided into a CPU
        `CPUNumpyArrayFrameResource` that holds a
        `np.ndarray`.
        """
        from yta_editor_frame.utils import _texture_to_numpy_frame

        if FrameResourceIdentifier.is_cpu(resource):
            # [1] `np.ndarray` -> `np.ndarray`
            return resource
        
        frame = None

        if FrameResourceIdentifier.is_cpu_av_videoframe(resource):
            # [2] `av.VideoFrame` -> `np.ndarray`
            import numpy as np

            numpy_array = resource.frame.to_ndarray(format = FORMAT)

            frame = numpy_array
        else:
            # [3] `moderngl.Texture` -> `np.ndarray`
            # [4] `render_target` -> `np.ndarray`

            # [3] `moderngl.Texture` -> `np.ndarray`
            texture = resource.frame
            if FrameResourceIdentifier.is_gpu_render_target(resource):
                # [4] `render_target` -> `np.ndarray`
                texture = texture.texture

            # `moderngl.Texture` -> `np.ndarray`
            numpy_array = _texture_to_numpy_frame(
                texture = texture,
                do_flip_texture = True,
                do_force_contiguous = False
            )

            frame = numpy_array

            # 2. Force 'rgba' format
            # TODO: Is this needed (?)
            # frame = _ensure_rgba_numpy_array(
            #     ndarray = numpy_array,
            #     # TODO: What do we do with the flip (?)
            #     # do_force_contiguous = do_force_contiguous
            #     do_force_contiguous = False
            # )

        return CPUNumpyArrayFrameResource(
            frame = frame,
            render_context = self.render_context
        )
        
    @requires_dependency('numpy', 'yta_editor_frame', 'numpy')
    def to_texture(
        self,
        resource: 'FrameResource'
    ) -> GPUTextureFrameResource:
        """
        Transform the `resource` provided into a GPU
        `GPUTextureFrameResource` that holds a
        `moderngl.Texture`.
        """
        from yta_editor_frame.utils import _numpy_frame_to_texture

        if FrameResourceIdentifier.is_gpu(resource):
            # [1] `moderngl.Texture` -> `moderngl.Texture`
            return resource
    
        frame = None
        
        if FrameResourceIdentifier.is_gpu_render_target(resource):
            # [2] `render_target` -> `moderngl.Texture`
            frame = copy_texture(
                texture = resource.frame.texture,
                render_context = self.render_context,
                do_get_from_pool = True
            )
        elif (
            FrameResourceIdentifier.is_cpu(resource) or
            FrameResourceIdentifier.is_cpu_av_videoframe(resource)
        ):
            # [3] `np.ndarray` -> `moderngl.Texture`
            # [4] `av.VideoFrame` -> `moderngl.Texture`
            numpy_array = resource.frame
            if FrameResourceIdentifier.is_cpu_av_videoframe(resource):
                # [4] `av.VideoFrame` -> `moderngl.Texture`
                numpy_array = numpy_array.to_ndarray(format = FORMAT)

            texture = _numpy_frame_to_texture(
                numpy_frame = numpy_array,
                render_context = self.render_context,
                do_flip = True,
                do_from_pool = True
            )

            frame = texture

            # We preserve the information about the alpha
            # channel that we had when numpy array

        return GPUTextureFrameResource(
            frame = frame,
            render_context = self.render_context
        )
    
    @requires_dependency('av', 'yta_editor_frame', 'av')
    def to_videoframe(
        self,
        resource: 'FrameResource'
    ) -> AvVideoFrameResource:
        """
        Transform the `resource` provided into a CPU
        `AvVideoFrameResource` that holds a
        `av.VideoFrame`.
        """
        from yta_editor_frame.utils import _texture_to_numpy_frame

        import av
        
        if FrameResourceIdentifier.is_cpu_av_videoframe(resource):
            # [1] `av.VideoFrame` -> `av.VideoFrame`
            return resource
        
        frame = None

        if FrameResourceIdentifier.is_cpu(resource):
            # [2] `np.ndarray` -> `av.VideoFrame`
            from yta_editor_frame.utils import _ensure_rgba_numpy_array

            # 1. Force 'rgba' format
            numpy_array = _ensure_rgba_numpy_array(
                ndarray = resource.frame,
                do_force_contiguous = True
            )

            # 2. `np.ndarray` -> `av.VideoFrame`
            frame = av.VideoFrame.from_ndarray(numpy_array, format = FORMAT)

            # We preserve the information about the alpha
            # channel that we had when numpy array
        else:
            # [3] `moderngl.Texture` -> `av.VideoFrame`
            # [4] `render_target` -> `av.VideoFrame`
            frame = resource.frame
            if FrameResourceIdentifier.is_gpu_render_target(resource):
                # [4] `render_target` -> `av.VideoFrame`
                frame = frame.texture

            # [3] `moderngl.Texture` -> `av.VideoFrame`
            # 1. `moderngl.Texture` -> `np.ndarray`
            numpy_array = _texture_to_numpy_frame(
                texture = frame,
                do_flip_texture = True,
                do_force_contiguous = True
            )

            # 2. `np.ndarray` -> `av.VideoFrame`
            frame = av.VideoFrame.from_ndarray(numpy_array, format = FORMAT)
        
        return AvVideoFrameResource(
            frame = frame,
            render_context = self.render_context
        )
    
    @requires_dependency('numpy', 'yta_editor_frame', 'numpy')
    def to_render_target(
        self,
        resource: 'FrameResource'
    ) -> RenderTargetFrameResource:
        """
        Transform the `resource` provided into a GPU
        `RenderTargetFrameResource` that holds a
        `RenderTarget`.
        """
        from yta_editor_frame.utils import _numpy_frame_to_texture

        if FrameResourceIdentifier.is_gpu_render_target(resource):
            # [1] `RenderTarget` -> `RenderTarget`
            return resource
        
        frame = None

        if FrameResourceIdentifier.is_gpu(resource):
            # [2] `moderngl.Texture` -> `RenderTarget`
            # Create a copy of the texture so the previous
            # owner can still have it and use it if needed
            texture = copy_texture(
                texture = resource.frame,
                render_context = self.render_context,
                do_get_from_pool = True
            )

            render_target = self.render_context.gpu.render_target.acquire(
                width = texture.width,
                height = texture.height,
                number_of_components = texture.components,
                dtype = texture.dtype
            )
            render_target.texture = frame
            
            frame = render_target
        elif (
            FrameResourceIdentifier.is_cpu(resource) or
            FrameResourceIdentifier.is_cpu_av_videoframe(resource)
        ):
            # [3] `np.ndarray` -> `RenderTarget`
            # [4] `av.VideoFrame` -> `RenderTarget`

            numpy_array = resource.frame
            if FrameResourceIdentifier.is_cpu_av_videoframe(resource):
                # [4] `av.VideoFrame` -> `RenderTarget`
                numpy_array = numpy_array.to_ndarray(format = FORMAT)

            # [3] `np.ndarray` -> `RenderTarget`
            # 1. Force 'rgba' format
            # TODO: Maybe this is necessary
            # numpy_array = _ensure_rgba_numpy_array(
            #     ndarray = numpy_array,
            #     do_force_contiguous = False
            # )

            # 2. `np.ndarray` -> `RenderTarget`
            frame = _numpy_frame_to_texture(
                numpy_frame = numpy_array,
                render_context = self.render_context,
                do_flip = True,
                do_from_pool = True
            )

            render_target = self.render_context.gpu.render_target.acquire(
                width = texture.width,
                height = texture.height,
                number_of_components = texture.components,
                dtype = texture.dtype
            )
            render_target.texture = frame

            # render_target.write(frame)

            frame = render_target

            # We preserve the information about the alpha
            # channel that we had when numpy array

        return RenderTargetFrameResource(
            frame = frame,
            render_context = self.render_context
        )


